/* Xmsgd 
 * Copyright (C) 2001 Thomas Vander Stichele <thomas@apestaart.org>
 * 
 * Xmsgd.c: main Xmsg daemon
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * version 0.6.4
 *
 * last changed : 28/08/2001
 *
 * displays a graphical message which disappears over time
 * also displays an .xpm icon in a 50x50 window
 * 
 * messages can be registered either through a fifo or a socket
 * using xml (not yet implemented)
 *
 */

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <glib.h>

/* project includes */

#include <config.h>
#include "cfg.h"
#include "../lib/default.h"
#include "../lib/Xmsglib.h"
#include "fifo.h"

/* default values */

/* globals */

cfg_t cfg;

#define MAX_LINES 5

/* function prototypes */

void            usage                   (const char *text);
GList*          get_line_msgp           (GList* msgs, int line_number);
Display*        open_display            (const char* display_name);
void            set_kill_time           (msg_t* msg);
void            display_msg             (Display* display, msg_t* msg);
void            check_kill_times        (GList** msgs, Display* display);

int main (int argc, char *argv[])
{
  Display* display;                     /* pointer to X Display structure */
  char *display_name = getenv("DISPLAY");  	/* address of the X display */
  GList *msgs = NULL;		/* GList to keep track of displayed messages */
  GList *msgp;
  msg_t *msg = NULL;
  //msg_t *msgs[MAX_LINES];			/* message structure array */

  /* fork the thing */
  if (fork ())
  {
    //printf ("Main program forks, exiting.\n");
    return 0;
  }  

  display = open_display (display_name);        /* open connection to X */

  /* initialize listener */
  strcpy (cfg.fifo, XMSGD_FIFO); //FIXME
  start_listener (cfg);

  /* this is where we should go in a loop */

  while (1)	/* is 1 a good choice ? */
  {
    // check if there's a message waiting
    MYDEBUG (("is there a message waiting ?\n"));
    if (message_waiting (cfg, &msg) > 0)
    {
      MYDEBUG ((" message waiting for line %d\n", msg->line_number));
      msgp = get_line_msgp (msgs, msg->line_number);
      if (msgp != NULL)
      {
        /* this line is already taken, kill it ! */
        KillMsg (display, msgp->data);
        FreeMsg (msgp->data);
        free (msgp->data);
        /* and remove it from the list */
        msgs = g_list_remove (msgs, msgp->data);
      }
      /* calculate kill time for this new message */
      set_kill_time (msg);
      MYDEBUG ((" msg font: %s\n", msg->font_name));
      msgs = g_list_append (msgs, msg);
      
      display_msg (display, msg);
      msg = NULL;
    }
    /* now do some housekeeping */
    
    check_kill_times (&msgs, display);
    MYDEBUG (("number of msgs on display: %d\n", g_list_length (msgs)));
    usleep (100000);
  }
  /* close the connection to the X server. */
  XCloseDisplay (display);

  return 0;
}

void 
usage (const char *text)
{
  printf ("%s\n\n", text);
  printf ("Usage info :\n\n");
/*
  printf ("Xmsgd \t[-i (icon .xpm)] [-c (X color)] [-l (line number)]\n");
  printf ("     \t[-d (duration)] [-f (X font descr)] (text to display)\n");
*/
}

GList*
get_line_msgp (GList* msgs, int line_number)
/* find the member of the glist that has this line number
 * if there isn't one, return NULL
 */
{
  int i;
  msg_t* msgp = NULL;
  for (i = 0; i < g_list_length (msgs); ++i)
  {
    msgp = (msg_t *) g_list_nth_data (msgs, i);
    if (msgp->line_number == line_number)
    {
      MYDEBUG (("returning msg with text %s\n", msgp->text));  
      return g_list_nth (msgs, i);
    }
  }
  return NULL;
}

Display* 
open_display (const char* display_name)
/* initially just meant to open display at start-up
 * now we want it to wait until we have a connection
 */
{
  Display* display;
  
  /* open connection with the X server */
  display = XOpenDisplay (display_name);
  while (display == NULL) 
  {
    sleep (NOX_SLEEP_TIME);
    MYDEBUG(("open_display: cannot connect to X server '%s'\n", display_name));
    display = XOpenDisplay (display_name);
  }
  return display;
}

void
set_kill_time (msg_t* msg)
{
  struct timeval tv;
  struct timezone tz;
  struct timeval kt;
  double duration = 0.0;
  long dur_usec = 0;
  
  /* get current timestamp, add duration, save */
  tz.tz_minuteswest = 0;
  tz.tz_dsttime = 0;
  duration = msg->duration;
  
  dur_usec = (long) (duration * (long) 1E6);
  
  if (gettimeofday (&tv, &tz) == -1)
  {
    printf ("ERROR: gettimeofday failed\n");
  }
  MYDEBUG (("set_kill_time: curtime %ld, %ld\n", tv.tv_sec, tv.tv_usec));
  kt.tv_usec = tv.tv_usec + (dur_usec % (long) 1E6);
  kt.tv_sec =  tv.tv_sec + dur_usec / (long) 1E6;
  if (kt.tv_usec >= (long) 1E6)
  {
    kt.tv_sec += kt.tv_usec / (long) 1E6;
    kt.tv_usec %= (long) 1E6;
  }
  msg->kill_time.tv_sec  = kt.tv_sec;
  msg->kill_time.tv_usec = kt.tv_usec;
  MYDEBUG (("set_kill_time: kt set to %ld, %ld\n", kt.tv_sec, kt.tv_usec));
}

long 
until_kill_time (msg_t* msg)
/* return 0 if we're past kill time, or return the number of usecs until kill time */
//FIXME: maybe we should add a margin of error here, the function takes time doesn't it ?
{  
  struct timeval tv;
  struct timezone tz;
  long diff_usec = 0;
  
  tz.tz_minuteswest = 0;
  tz.tz_dsttime = 0;
  
  if (gettimeofday (&tv, &tz) == -1)
  {
    printf ("ERROR: gettimeofday failed\n");
  }
  MYDEBUG (("until_kill_time: curtime %ld, %ld\n", tv.tv_sec, tv.tv_usec));
  /* first check the cases were we should return 0 */
  if (msg->kill_time.tv_sec < tv.tv_sec)
    return 0L;
  if ((msg->kill_time.tv_sec == tv.tv_sec)
   && (msg->kill_time.tv_usec <= tv.tv_usec))
    return 0L;
  /* we haven't reached it yet */
  diff_usec = (msg->kill_time.tv_sec - tv.tv_sec) * (long) 1E6
            + (msg->kill_time.tv_usec - tv.tv_usec);
  MYDEBUG (("until_kill_time: %ld usecs left\n", diff_usec));
  return diff_usec;
}

void 
display_msg (Display* display, msg_t* msg)
/* do necessary X stuff and show the msg on-screen */
{
  int screen_num;               /* number of screen to place the window on */
  int x, y;
  
  MYDEBUG (("loading font %s\n", msg->font_name));
  LoadFont (display, msg);

  /* calculate text width */
  msg->text_width = XTextWidth (msg->font_info, msg->text,
  strlen (msg->text)) + 4 /* shadows */;

  /* get the geometry of the default screen for our display */
  XFlush (display);
  screen_num = DefaultScreen (display);
  msg->display_width = DisplayWidth (display, screen_num);
  msg->display_height = DisplayHeight (display, screen_num);

  /* take lowest value of either display_width - IMG_W or text_width */
  if (msg->display_width - IMG_W > msg->text_width)
  msg->message_width = msg->text_width;
  else
  msg->message_width = msg->display_width - IMG_W;

  /* calculate geometries for window */
  if (msg->random)
  {
    Screen* screen;
    int max_x, max_y;

    screen = ScreenOfDisplay (display, screen_num);
    max_x = WidthOfScreen (screen);
    max_y = HeightOfScreen (screen);
    x = (rand () / (RAND_MAX * 1.0)) * (max_x - msg->message_width);
    y = (rand () / (RAND_MAX * 1.0)) * max_y;
  }
  else
  {
    x = 0;
    y = msg->line_number * MSG_H;
  }
  
  MYDEBUG ((" Xmsgd: creating windows\n")); 
  
  msg->img_w = CreateWindow (display, screen_num,  x, y, 
  IMG_W - 1, IMG_H - 1);
  msg->msg_w = CreateWindow (display, screen_num, x + 50, y, 
  msg->message_width, MSG_H - 1);

  XFlush (display);
  MYDEBUG ((" Xmsgd: windows created\n")); 

  DrawText (display, msg->msg_w, *msg, msg->text);
  if (msg->icon_file != NULL)
  {
    DrawMsgXpm (display, msg->img_w, *msg);
  }

  XFlush (display);
}

void
check_kill_times (GList **msgs, Display* display)
/* walk through a list with for *backwards*, and kill every message that's
 * past it's kill time
 */
{
  int i;
  
  MYDEBUG (("checking kill times\n"));
  MYDEBUG (("%d msgs on screen\n", g_list_length (*msgs)));

  if (g_list_length (*msgs) == 0)
  {
    MYDEBUG (("no elements in list, returning\n"));
    return;
  }
  for (i = g_list_length (*msgs); i > 0; --i)
  {
    msg_t* msg = NULL;
    msg = (msg_t*) g_list_nth_data (*msgs, i - 1);
    if (!msg) { printf ("ERROR: could not get %d'th element\n", i - 1); }
    MYDEBUG (("i=%d, msg line=%d\n", 
            i, msg->line_number));
    if (until_kill_time (msg) == 0)
    {
      MYDEBUG (("killing msg on line %d\n", msg->line_number));
      *msgs = g_list_remove (*msgs, msg);
      KillMsg (display, msg);
      FreeMsg (msg);
      free (msg);
  MYDEBUG (("%d msgs on screen\n", g_list_length (*msgs)));
    }  
  }
}
